import { Kind } from '../language/kinds.mjs';
import { visit } from '../language/visitor.mjs';
import { getNamedType, isNamedType } from '../type/definition.mjs';
import { TypeInfo, visitWithTypeInfo } from './TypeInfo.mjs';
export function mergeAST(documentAST, schema) {
  // If we're given the schema, we can simplify even further by resolving object
  // types vs unions/interfaces
  const typeInfo = schema ? new TypeInfo(schema) : null;
  const fragmentDefinitions = Object.create(null);

  for (const definition of documentAST.definitions) {
    if (definition.kind === Kind.FRAGMENT_DEFINITION) {
      fragmentDefinitions[definition.name.value] = definition;
    }
  }

  const flattenVisitors = {
    SelectionSet(node) {
      const selectionSetType =
        typeInfo === null || typeInfo === void 0
          ? void 0
          : typeInfo.getParentType();
      let { selections } = node;
      selections = selectionSetType
        ? inlineRelevantFragmentSpreads(
            fragmentDefinitions,
            selections,
            selectionSetType,
          )
        : inlineRelevantFragmentSpreads(fragmentDefinitions, selections);
      return { ...node, selections };
    },

    FragmentDefinition() {
      return null;
    },
  };
  const flattenedAST = visit(
    documentAST,
    typeInfo ? visitWithTypeInfo(typeInfo, flattenVisitors) : flattenVisitors,
  );
  const deduplicateVisitors = {
    SelectionSet(node) {
      let { selections } = node;
      selections = uniqueBy(selections, (selection) =>
        selection.alias ? selection.alias.value : selection.name.value,
      );
      return { ...node, selections };
    },

    FragmentDefinition() {
      return null;
    },
  };
  return visit(flattenedAST, deduplicateVisitors);
}

function inlineRelevantFragmentSpreads(
  fragmentDefinitions,
  selections,
  selectionSetType,
) {
  // const selectionSetTypeName = selectionSetType
  //   ? getNamedType(selectionSetType).name
  //   : null;
  let selectionSetTypeName = null;

  if (selectionSetType) {
    const typ = getNamedType(selectionSetType);

    if (isNamedType(typ)) {
      selectionSetTypeName = typ.name;
    }
  }

  const outputSelections = [];
  const seenSpreads = [];

  for (let selection of selections) {
    var _selection$directives;

    if (selection.kind === 'FragmentSpread') {
      const fragmentName = selection.name.value;

      if (!selection.directives || selection.directives.length === 0) {
        if (seenSpreads.includes(fragmentName)) {
          /* It's a duplicate - skip it! */
          continue;
        } else {
          seenSpreads.push(fragmentName);
        }
      }

      const fragmentDefinition = fragmentDefinitions[selection.name.value];

      if (fragmentDefinition) {
        const { typeCondition, directives, selectionSet } = fragmentDefinition;
        selection = {
          kind: Kind.INLINE_FRAGMENT,
          typeCondition,
          directives,
          selectionSet,
        };
      }
    }

    if (
      selection.kind === Kind.INLINE_FRAGMENT && // Cannot inline if there are directives
      (!selection.directives ||
        ((_selection$directives = selection.directives) === null ||
        _selection$directives === void 0
          ? void 0
          : _selection$directives.length) === 0)
    ) {
      const fragmentTypeName = selection.typeCondition
        ? selection.typeCondition.name.value
        : null;

      if (!fragmentTypeName || fragmentTypeName === selectionSetTypeName) {
        outputSelections.push(
          ...inlineRelevantFragmentSpreads(
            fragmentDefinitions,
            selection.selectionSet.selections,
            selectionSetType,
          ),
        );
        continue;
      }
    }

    outputSelections.push(selection);
  }

  return outputSelections;
}

function uniqueBy(array, iteratee) {
  const FilteredMap = new Map();
  const result = [];

  for (const item of array) {
    if (item.kind === 'Field') {
      var _item$directives;

      const uniqueValue = iteratee(item);
      const existing = FilteredMap.get(uniqueValue);

      if (
        (_item$directives = item.directives) !== null &&
        _item$directives !== void 0 &&
        _item$directives.length
      ) {
        // Cannot inline fields with directives (yet)
        const itemClone = { ...item };
        result.push(itemClone);
      } else if (
        existing !== null &&
        existing !== void 0 &&
        existing.selectionSet &&
        item.selectionSet
      ) {
        // Merge the selection sets
        existing.selectionSet.selections = [
          ...existing.selectionSet.selections,
          ...item.selectionSet.selections,
        ];
      } else if (!existing) {
        const itemClone = { ...item };
        FilteredMap.set(uniqueValue, itemClone);
        result.push(itemClone);
      }
    } else {
      result.push(item);
    }
  }

  return result;
}
